/*
Author: Gao Mingcen & Cao Thanh Tung
Date: 28/02/2013

File Name: FFHull_Basic.cpp

Basic work flow of FFHull, see the funcion ComputeConvexHull();

===============================================================================

Copyright (c) 2012, 2013, School of Computing, National University of Singapore. 
All rights reserved.

Project homepage: http://www.comp.nus.edu.sg/~tants/flipflop.html

If you use ffHull and you like it or have comments on its usefulness etc., we 
would love to hear from you at <tants@comp.nus.edu.sg>. You may share with us
your experience and any possibilities that we may improve the work/code.

===============================================================================

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the National University of Singapore nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission from the National University of Singapore. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

#include "FFHull.h"
#include "PerformanceTimer.h"

#include <math.h>

FFHull::FFHull(TriangleMesh3D * tMesh)
{
	this->cmesh = tMesh;
	this->mode = -1; //
	this->growing = false;
}

FFHull::~FFHull(void)
{
	this->cmesh = NULL;
	this->mode = -1;
}

void FFHull::_setMode(int _mode)
{
	if(_mode != 0 && _mode!= 1)
	{
		printf("FFHull Error: Wrong Mode %d (0:CPU 1:GPU)\n", _mode);
		exit(0);
	}
	this->mode = _mode;
}
void FFHull::_unsetMode()
{
	this->mode = -1;
}
TriangleMesh3D * FFHull::ComputeConvexHull()
{
	return this->ComputeConvexHull(0);
}
TriangleMesh3D * FFHull::ComputeConvexHull(int _mode)
{
	return this->ComputeConvexHull(_mode, false);
}
TriangleMesh3D * FFHull::ComputeConvexHull(int _mode, bool _detail)
{
	if(this->cmesh->noPoint <=4) 
	{
		printf("FFHull: trival case\n");
		exit(0);
	}
	//0. Initialization
	this->detail = _detail;
	this->_setMode(_mode);
	this->Initialization();
	if(detail) printf("**0. [CH3D-%d] Initialization\n", this->mode);

	//1. Compute the tetrahedral
	this->ComputeTetra();
	if(detail) printf("**1. [CH3D-%d] Compute Tetrahedra\n", this->mode);
	
	//2. Begin the loop
	int count = 0; 
	while(growing)
	{
		this->GrowHull();
		count ++;
	}
	if(detail) printf("**2. [CH3D-%d] Assign %d Layers\n", this->mode, count);

	//3. Correct the hull
	this->CorrectHull();
	if(detail) printf("**3. [CH3D-%d] PFlipping to Fix\n", this->mode);

	//4. compact the triangle cmesh
	this->CompactMesh();
	if(detail) printf("**4. [CH3D-%d] Compact Mesh\n", this->mode);

	//5. deinitialization
	this->Deinitialization();
	if(detail) printf("**5. [CH3D-%d] Deinitialization\n", this->mode);

	return this->cmesh;
}

void FFHull::Initialization()
{
	if(this->mode == 0)
		c_Initialization();
	else if(this->mode == 1)
		g_Initialization();
}

void FFHull::Deinitialization()
{
	if(this->mode == 0)
		c_Deinitialization();
	else if(this->mode == 1)
		g_Deinitialization();
}

void FFHull::ComputeTetra()
{	
	if(this->mode == 0)
		c_ComputeTetra();
	else if(this->mode == 1)
		g_ComputeTetra();
}
void FFHull::GrowHull()
{
	if(this->mode == 0)
		c_GrowHull();
	else if(this->mode == 1)
		g_GrowHull();
}
void FFHull::CorrectHull()
{
	if(this->mode == 0)
		c_CorrectHull();
	else if(this->mode == 1)
		g_CorrectHull();
}

void FFHull::CompactMesh()
{
	if(this->mode == 0)
		c_CompactMesh();
	else if(this->mode == 1)
		g_CompactMesh();
}
