/*
Author: Gao Mingcen & Cao Thanh Tung
Date: 28/02/2013

File Name: FFHull_CPU.cpp

Functions of FFHull in the CPU version

===============================================================================

Copyright (c) 2012, 2013, School of Computing, National University of Singapore. 
All rights reserved.

Project homepage: http://www.comp.nus.edu.sg/~tants/flipflop.html

If you use ffHull and you like it or have comments on its usefulness etc., we 
would love to hear from you at <tants@comp.nus.edu.sg>. You may share with us
your experience and any possibilities that we may improve the work/code.

===============================================================================

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the National University of Singapore nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission from the National University of Singapore. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

#include <math.h>
#include "FFHull.h"
#include "cpuPredicates.h"

void FFHull::c_Initialization()
{
	//exact computation
	this->c_exactInit();

	if(cmesh->noTri == 0)
	{
		this->triArraySize = max(cmesh->noPoint/2, 4);
		this->cmesh->triangles = (Triangle *) malloc(this->triArraySize*sizeof(Triangle));
	}
	else
	{
		this->triArraySize = cmesh->noTri;
	}
	this->triToPoint = (int *) malloc(this->triArraySize*sizeof(int));
	this->triToValue = (REAL *) malloc(this->triArraySize*sizeof(REAL));
    this->pointOrder = (int *) malloc(this->cmesh->noPoint * sizeof(int)); 
    this->pointLoc = (char *) malloc(this->cmesh->noPoint * sizeof(char)); 
}

void FFHull::c_Deinitialization()
{
	free(this->triToPoint);
	free(this->triToValue);
	free(this->pointOrder);
	free(this->pointLoc);
}

void FFHull::c_ComputeTetra()
{
	//1. find extreme
	REAL maxxyz, maxx_y_z, max_xy_z, max_x_yz;
	maxxyz= maxx_y_z= max_xy_z= max_x_yz= MINREAL;
	for(int i = 0; i< cmesh->noPoint; i++)
	{
		REAL value = cmesh->points[i].x +  cmesh->points[i].y + cmesh->points[i].z;
		if(value > maxxyz)
		{
			maxxyz = value;
			iniPoint[0] = i;
		}
	}
	for(int i = 0; i< cmesh->noPoint; i++)
	{
		if(i == iniPoint[0]) continue;
		REAL value = cmesh->points[i].x -  cmesh->points[i].y - cmesh->points[i].z;
		if(value > maxx_y_z)
		{
			maxx_y_z = value;
			iniPoint[1] = i;
		}
	}
	for(int i = 0; i< cmesh->noPoint; i++)
	{
		if(i == iniPoint[0] || i == iniPoint[1]) continue;
		REAL value = -cmesh->points[i].x +  cmesh->points[i].y - cmesh->points[i].z;
		if(value > max_xy_z)
		{
			max_xy_z = value;
			iniPoint[2] = i;
		}
	}
	for(int i = 0; i< cmesh->noPoint; i++)
	{
		if(i == iniPoint[0] || i == iniPoint[1] || i == iniPoint[2]) continue;
		REAL value = -cmesh->points[i].x -  cmesh->points[i].y + cmesh->points[i].z;
		if(value > max_x_yz)
		{
			max_x_yz = value;
			iniPoint[3] = i;
		}
	}

	//2. construct kernel
	cmesh->kernel.x = cmesh->kernel.y = cmesh->kernel.z = 0;
	for(int i =0; i< 4; i++)
	{
		cmesh->kernel.x += cmesh->points[iniPoint[i]].x;
		cmesh->kernel.y += cmesh->points[iniPoint[i]].y;
		cmesh->kernel.z += cmesh->points[iniPoint[i]].z;
	}
	cmesh->kernel.x/=4.0;
	cmesh->kernel.y/=4.0;
	cmesh->kernel.z/=4.0;

	//3. construct terahedral
	REAL res = this->c_check3DOrientation(cmesh, iniPoint[0], iniPoint[1], iniPoint[2], iniPoint[3]);
	if(res > 0)
	{
		int temp = iniPoint[0];
		iniPoint[0] = iniPoint[1];
		iniPoint[1] = temp;
	}
	cmesh->noTri = 4;
	cmesh->triangles[0].vtx[0] = iniPoint[0];
	cmesh->triangles[0].vtx[1] = iniPoint[1];
	cmesh->triangles[0].vtx[2] = iniPoint[2];
	cmesh->triangles[1].vtx[0] = iniPoint[2];
	cmesh->triangles[1].vtx[1] = iniPoint[1];
	cmesh->triangles[1].vtx[2] = iniPoint[3];
	cmesh->triangles[2].vtx[0] = iniPoint[0];
	cmesh->triangles[2].vtx[1] = iniPoint[2];
	cmesh->triangles[2].vtx[2] = iniPoint[3];
	cmesh->triangles[3].vtx[0] = iniPoint[1];
	cmesh->triangles[3].vtx[1] = iniPoint[0];
	cmesh->triangles[3].vtx[2] = iniPoint[3];
	cmesh->triangles[0].neighbor[0] = Encode(1,2);
	cmesh->triangles[0].neighbor[1] = Encode(2,2);
	cmesh->triangles[0].neighbor[2] = Encode(3,2);

	cmesh->triangles[1].neighbor[0] = Encode(3,1);
	cmesh->triangles[1].neighbor[1] = Encode(2,0);
	cmesh->triangles[1].neighbor[2] = Encode(0,0);

	cmesh->triangles[2].neighbor[0] = Encode(1,1);
	cmesh->triangles[2].neighbor[1] = Encode(3,0);
	cmesh->triangles[2].neighbor[2] = Encode(0,1);

	cmesh->triangles[3].neighbor[0] = Encode(2,1);
	cmesh->triangles[3].neighbor[1] = Encode(1,0);
	cmesh->triangles[3].neighbor[2] = Encode(0,2);

	//4. initialize triToPoint, triToValue; and cmesh->pointMap
	for(int i = 0; i< cmesh->noTri; i++)
	{		
		this->triToPoint[i] = -1; //means the point assign in next round
		this->triToValue[i] = MINREAL; //initial value
	}
	cmesh->pointMap[iniPoint[0]] = -2;
	cmesh->pointMap[iniPoint[1]] = -2;
	cmesh->pointMap[iniPoint[2]] = -2;
	cmesh->pointMap[iniPoint[3]] = -2;

	//5. Walking to find the correct group

	this->c_correctInitialGroup();
}
void FFHull::c_GrowHull()
{
    if ( cStack.empty() ) 
    {
        this->growing = false; 
        return ; 
    }

    StackItem item = cStack.back(); 
    cStack.pop_back(); 

	//1. Assign the layer
	if (this->c_assignLayer(item)) 
	{
		//2. Subdivide the point, and set growing to be true if necessary
		this->c_subdivide(item);
	}
}
void FFHull::c_CorrectHull()
{
	//3. Fix the Hull, may use filpping and retrigulating
	// Push all edges to the queue
    for ( int i = 0; i < this->cmesh->noTri; i++ ) 
        for ( int j = 0; j < 3; j++ ) 
            if ( i < Tri(this->cmesh->triangles[i].neighbor[j]) )
            {
                cCheckEdge.push_back(Edge(Encode(i,j), Tri(this->cmesh->triangles[i].neighbor[j])));
            }  

	int count = 0;
	//fix all edges
	while(!cCheckEdge.empty())
	{
        const Edge e = cCheckEdge.back(); 
		cCheckEdge.pop_back();
		this->c_pFlipEdge(e);
		
		count++;
	}
	if(detail) printf("**** Check %d edges in total.\n", count);
}
void FFHull::c_CompactMesh()
{
	int oldTriNum = this->cmesh->noTri;
	int count = 0;
	int * newPos = new int[this->cmesh->noTri];
	for(int i = 0; i< this->cmesh->noTri; i++)
	{
		Triangle * tri = & this->cmesh->triangles[i];
		if(tri->vtx[0] == -1)
		{
			newPos[i] = -1;
		}
		else
		{
			newPos[i] = count;
			count++;
		}
	}
	for(int i = 0; i< this->cmesh->noTri; i++)
	{
		int tid = newPos[i];
		if(tid == -1) continue;
		Triangle * tri = & this->cmesh->triangles[i];
		for(int j = 0; j< 3; j++)
		{
			this->cmesh->triangles[tid].vtx[j] = tri->vtx[j];
			this->cmesh->triangles[tid].neighbor[j] = Encode(newPos[Tri(tri->neighbor[j])], Pos(tri->neighbor[j]));
			
			//update pointmap
			this->cmesh->pointMap[tri->vtx[j]] = tid; //for test
		}
	}
	this->cmesh->noTri = count;
	delete[] newPos;
	if(detail) printf("**** Compact triangles from %d to %d\n", oldTriNum, count);
}

bool FFHull::c_assignLayer(const StackItem item)
{
	Triangle * tris[3];
	int triIds[3], oppoIds[3];
	int oldNoTri = this->cmesh->noTri;

    int i = item.triangle; 
	if(this->triToPoint[i] == -1)
        return false; 

	//expand array if necessary
	int curTriNo = this->cmesh->noTri;
	if(this->triArraySize < curTriNo + 2)
	{
		int newSize = (this->cmesh->noTri +2)*2;
		this->c_expandArray((void **)&this->cmesh->triangles, curTriNo * sizeof(Triangle), newSize*sizeof(Triangle));
		this->c_expandArray((void **)&this->triToPoint, curTriNo * sizeof(int), newSize*sizeof(int));
		this->c_expandArray((void **)&this->triToValue, curTriNo * sizeof(REAL), newSize*sizeof(REAL));
		this->triArraySize = newSize;
		printf("Expand\n");
	}
	
	//basic info
	int pid = this->triToPoint[i];
	Triangle * thisTri = &this->cmesh->triangles[i];

	//set pointMap
    this->cmesh->pointMap[pid] = -2; 

	//replace
	//new triangle with 2-0(oppo 1) and with 0-1(oppo 2)
	Triangle * tri20 = &this->cmesh->triangles[curTriNo];
	Triangle * tri01 = &this->cmesh->triangles[curTriNo + 1];
	tri20->vtx[0] = pid; tri20->vtx[1] = thisTri->vtx[2]; tri20->vtx[2] = thisTri->vtx[0];
	tri20->neighbor[0] = thisTri->neighbor[1]; tri20->neighbor[1]= Encode(curTriNo+1, 2); tri20->neighbor[2] = Encode(i, 1);
	tri01->vtx[0] = pid; tri01->vtx[1] = thisTri->vtx[0]; tri01->vtx[2] = thisTri->vtx[1];
	tri01->neighbor[0] = thisTri->neighbor[2]; tri01->neighbor[1]= Encode(i, 2); tri01->neighbor[2] = Encode(curTriNo, 1);
	int oppoTri20 = Tri(thisTri->neighbor[1]);
	int oppoPos20 = Pos(thisTri->neighbor[1]);
	this->cmesh->triangles[oppoTri20].neighbor[oppoPos20] = Encode(curTriNo, 0);
	int oppoTri01 = Tri(thisTri->neighbor[2]);
	int oppoPos01 = Pos(thisTri->neighbor[2]);
	this->cmesh->triangles[oppoTri01].neighbor[oppoPos01] = Encode(curTriNo+1, 0);

	thisTri->vtx[0] = pid; thisTri->neighbor[1] = Encode(curTriNo, 2); thisTri->neighbor[2] = Encode(curTriNo+1, 1);

	//update triToPoint, triToValue and pointToTri
	this->triToPoint[i] = -1; this->triToPoint[curTriNo] = -1; this->triToPoint[curTriNo + 1] = -1; 
	this->triToValue[i] = MINREAL; this->triToValue[curTriNo] = MINREAL; this->triToValue[curTriNo + 1] = MINREAL;

	this->cmesh->noTri +=2;
    return true; 
}

void FFHull::c_subdivide(const StackItem item)
{		
	int ps[3], map, offset, code, v1, v2, shift;
	REAL res;
	Triangle thisTri = this->cmesh->triangles[item.triangle];

	bool isBoundary;
    int p; 

	ps[0] = this->cmesh->triangles[Tri(thisTri.neighbor[1])].vtx[Pos(thisTri.neighbor[1])];
	ps[1] = thisTri.vtx[1];
	ps[2] = thisTri.vtx[2];

	for(p = item.fromPoint; p < item.toPoint; p++)
	{
        this->pointLoc[p] = -1; 
        int i = this->pointOrder[p]; 
		map = this->cmesh->pointMap[i];

        if(map >= 0)
		{
			res = this->c_check3DOrientation(this->cmesh, -1, thisTri.vtx[0], ps[0], i);
			//complex logic for degeneracy. If using SOS, it should be simpler
			if(res != 0.0)
			{
				code = res > 0? 1: 2;
				res = this->c_check3DOrientation(this->cmesh, -1, thisTri.vtx[0], ps[code], i);
				if(code == 1) offset = (res <= 0.0)?2: 0;
				else offset = (res >= 0.0)?1: 0;
			}
			else
			{
				res = this->c_check3DOrientation(this->cmesh, -1, ps[1], ps[2], thisTri.vtx[0]);
				if(res != 0.0)//offset ==0
				{					
					//test c2i
					res = this->c_check3DOrientation(this->cmesh, -1, thisTri.vtx[0], ps[2], i); 
					if(res != 0.0) offset = (res<0)?0:1;
					else
					{
						//test c1i
						res = this->c_check3DOrientation(this->cmesh, -1, thisTri.vtx[0], ps[1], i); 
						offset = (res>=0)?0:2;
					}
				}
				else
				{
					res = this->c_check3DOrientation(this->cmesh, -1, ps[0], ps[2], thisTri.vtx[0]);
					if(res == 0.0) offset = 2;
					else offset = 1;
				}
			}

			//check whether is dead
			res = this->c_check3DOrientation(this->cmesh, 
				thisTri.vtx[0], ps[Next(offset)], ps[Prev(offset)], i);
			if(res <= 0) {
				map = -1; //it is inside the new tetrahedra
			}
			else //record the furthest point
			{
				if(offset>0) map = Tri(thisTri.neighbor[offset]);
				if(res > this->triToValue[map])
				{
					this->triToValue[map] = res;
					this->triToPoint[map] = i;
				}
				this->pointLoc[p] = offset; 
			}
			this->cmesh->pointMap[i] = map;
		}
		//else the point is dead
	}

    int pos[4]; 
    int tri[3] = { item.triangle, Tri(thisTri.neighbor[1]), Tri(thisTri.neighbor[2]) }; 
    
    c_reorderPoints( this->pointOrder, this->pointLoc, pos, 3, item.fromPoint, item.toPoint ); 

    for ( int i = 0; i < 3; i++ ) 
        if ( pos[i] < pos[i+1] ) 
        {
            const StackItem newItem = { item.fromPoint + pos[i], item.fromPoint + pos[i+1], tri[i] }; 
            cStack.push_back(newItem); 
        }
}

void FFHull::c_pFlipEdge(const Edge edge)
{
	int triId1 = Tri(edge.a);
	int pos1 = Pos(edge.a);
	int triId2 = edge.b;
	int min12, minNP;
	bool toFlip;

	//check if the edge is existing
	if(triId1 < 0 || triId1 >= this->cmesh->noTri) return;
	if(triId2 < 0 || triId2 >= this->cmesh->noTri) return;
	if(pos1<0 || pos1>2) return;
	Triangle * tri1 = &this->cmesh->triangles[triId1];
	if(tri1->vtx[0] == -1) return;
	if(Tri(tri1->neighbor[pos1]) != triId2) return;

	//check reflexity of the edge
	Triangle * tri2 = &this->cmesh->triangles[triId2];
	if(tri2->vtx[0] == -1) return;
	int pos2 = Pos(tri1->neighbor[pos1]);
	int pid1 = tri1->vtx[pos1];
	int pid2 = tri2->vtx[pos2];
	int pidNext = tri1->vtx[Next(pos1)];
	int pidPrev = tri1->vtx[Prev(pos1)];

	REAL edgeRes = this->c_check3DOrientation(this->cmesh, pid1, pidNext, pidPrev, pid2);
	int offset = MAXINT;
	if(this->cmesh->pointMap[pid1] == -1 && pid1 < offset) offset = pid1;
	if(this->cmesh->pointMap[pid2] == -1 && pid2 < offset) offset = pid2;
	if(this->cmesh->pointMap[pidNext] == -1 && pidNext < offset) offset = -1;
	else if(this->cmesh->pointMap[pidPrev] == -1 && pidPrev < offset) offset = -1;

	if(edgeRes < 0 && offset != -1) return; //if convex and p-flip is not met, return;
	//<=0: convex. ==0 do nothing, otherwise flip back and for

	//if it is reflex, continue
	//check the topology.Because we keep star shape, so any 3-1 topology can be flippable.
	if(Tri(tri1->neighbor[Prev(pos1)]) == Tri(tri2->neighbor[Next(pos2)])) //3-1, and pidNext is in the cone
	{
		//this must be edgeRes>=0. Remove thin triangle
		if(edgeRes >= 0 || this->cmesh->pointMap[pidNext] == -1) //edge is reflex or pidNext degree is 3
		{
			this->c_flip31(edge.a, Encode(triId2, pos2));
		}
	}
	else if(Tri(tri1->neighbor[Next(pos1)]) == Tri(tri2->neighbor[Prev(pos2)])) //3-1, and pidPrev is in the cone
	{
		if(edgeRes >= 0 || this->cmesh->pointMap[pidPrev] == -1) // edge is reflex or pidPrev degree is 3
		{
			this->c_flip31(Encode(triId2, pos2), edge.a);
		}
	}
	else //if the topology is 2-2
	{
		//3 cases to do: reflex and unflippable(next and prev); do flipping
		//if kernel is outside the tetrahedral, the algorithm is clear
		//if kernel is inside the tetrahedral(edge must be convex and shape is <->), we quickly remove it
		REAL conRes = this->c_check3DOrientation(this->cmesh, -1, pid1, pid2, pidNext);
		if(conRes >= 0 && edgeRes >= 0 && this->cmesh->pointMap[pidNext]!=-1)
		// pidNext is inside the cone kernel-pid1-pid2-pidPrev + reflex ->unflippable
		{
			//label pidNext
			this->cmesh->pointMap[pidNext] = -1;
			//push all edge to check 
			int curTri = triId1;
			int curPos = Prev(pos1);
			int nextCode;
			while(curTri != triId2)
			{						
				nextCode = this->cmesh->triangles[curTri].neighbor[curPos];
				this->cCheckEdge.push_back(Edge(nextCode, curTri));
				curTri = Tri(nextCode);
				curPos = Prev(Pos(nextCode));
			}
		}
		else if(conRes <=0)
		{
			conRes = this->c_check3DOrientation(this->cmesh, -1, pid2, pid1, pidPrev);
			if(conRes >= 0 && edgeRes >= 0 && this->cmesh->pointMap[pidPrev]!=-1)
			//pidPrev is inside the conve kernel-pid2-pid1-pidNext + reflex -> unflippable
			{
				//label pidPrev
				this->cmesh->pointMap[pidPrev] = -1;
				//push all edge to check except 
				int curTri = triId2;
				int curPos = Prev(pos2);
				int nextCode;
				while(curTri != triId1)
				{				
					nextCode = this->cmesh->triangles[curTri].neighbor[curPos];
					this->cCheckEdge.push_back(Edge(nextCode, curTri));
					curTri = Tri(nextCode);
					curPos = Prev(Pos(nextCode));
				}
			}
			//may continue check pid1 and pid2
			else if(conRes <= 0)//flippable
			{
				if(offset == MAXINT) toFlip = (edgeRes > 0);
				else toFlip = (offset == -1);

				if(toFlip)
				{
					this->c_flip22(edge.a, Encode(triId2, pos2));
				}
			}
		}
	}
}
void FFHull::c_flip31(int triCodeLeft, int triCodeRight)
{
	int triId1 = Tri(triCodeLeft);
	int pos1 = Pos(triCodeLeft);
	int triId2 = Tri(triCodeRight);
	int pos2 = Pos(triCodeRight);
	Triangle * tri1 = &this->cmesh->triangles[triId1];
	Triangle * tri2 = &this->cmesh->triangles[triId2];

	int sharedTriId = Tri(tri1->neighbor[Prev(pos1)]);
	int sharedPos = Pos(tri1->neighbor[Prev(pos1)]);
	Triangle * sharedTri = & this->cmesh->triangles[sharedTriId];
	
	int pid = tri1->vtx[Next(pos1)];
	int pid1 = tri1->vtx[pos1];
	int pid2 = tri2->vtx[pos2];
	int pidOut = tri1->vtx[Prev(pos1)];

	//remove triId1, triId2 and sharedTriId
	//put the new triangle on triId1
	tri1->neighbor[pos1] = tri2->neighbor[Prev(pos2)];
	tri1->neighbor[Prev(pos1)] = sharedTri->neighbor[Next(sharedPos)];
	tri1->vtx[Next(pos1)] = pid2;
	this->cCheckEdge.push_back(Edge(tri1->neighbor[Next(pos1)], triId1));

	//update the neighbor
	int neiTriId = Tri(tri2->neighbor[Prev(pos2)]);
	int neiPos = Pos(tri2->neighbor[Prev(pos2)]);
	this->cmesh->triangles[neiTriId].neighbor[neiPos] = Encode(triId1, pos1);
	this->cCheckEdge.push_back(Edge(Encode(triId1, pos1),neiTriId));

	neiTriId = Tri(sharedTri->neighbor[Next(sharedPos)]);
	neiPos = Pos(sharedTri->neighbor[Next(sharedPos)]);
	this->cmesh->triangles[neiTriId].neighbor[neiPos] = Encode(triId1, Prev(pos1));
	this->cCheckEdge.push_back(Edge(Encode(triId1, Prev(pos1)),neiTriId));

	//"remove" tri2 and sharedTri
	this->c_clearTriangle(*tri2);
	this->c_clearTriangle(*sharedTri);

	//update pointMap
	this->cmesh->pointMap[pid] = -1;
}
void FFHull::c_flip22(int triCodeLeft, int triCodeRight)
{	
	int triId1 = Tri(triCodeLeft);
	int pos1 = Pos(triCodeLeft);
	int triId2 = Tri(triCodeRight);
	int pos2 = Pos(triCodeRight);
	Triangle * tri1 = &this->cmesh->triangles[triId1];
	Triangle * tri2 = &this->cmesh->triangles[triId2];
	
	int pid1 = tri1->vtx[pos1];
	int pid2 = tri2->vtx[pos2];
	int pidPrev = tri1->vtx[Prev(pos1)];
	int pidNext = tri1->vtx[Next(pos1)];

	int nei1NextCode = tri1->neighbor[Next(pos1)];
	int nei1PrevCode = tri1->neighbor[Prev(pos1)];
	int nei2NextCode = tri2->neighbor[Next(pos2)];
	int nei2PrevCode = tri2->neighbor[Prev(pos2)];

	//update the neighbor
	int neiTriId = Tri(nei1PrevCode);
	int neiPos = Pos(nei1PrevCode);
	this->cmesh->triangles[neiTriId].neighbor[neiPos] = Encode(triId2, pos2);
	this->cCheckEdge.push_back(Edge(Encode(triId2, pos2), neiTriId));

	neiTriId = Tri(nei2PrevCode);
	neiPos = Pos(nei2PrevCode);
	this->cmesh->triangles[neiTriId].neighbor[neiPos] = Encode(triId1, pos1);
	this->cCheckEdge.push_back(Edge(Encode(triId1, pos1), neiTriId));

	//update tri1 and tri2
	tri1->vtx[Next(pos1)] = pid2;
	tri2->vtx[Next(pos2)] = pid1;
	tri1->neighbor[pos1] = nei2PrevCode;
	tri1->neighbor[Prev(pos1)] = Encode(triId2, Prev(pos2));
	tri2->neighbor[pos2] = nei1PrevCode;
	tri2->neighbor[Prev(pos2)] = Encode(triId1, Prev(pos1));

	this->cCheckEdge.push_back(Edge(nei1NextCode, triId1));
	this->cCheckEdge.push_back(Edge(nei2NextCode, triId2));
}

void FFHull::c_correctInitialGroup()
{
	REAL res;
    memset(this->pointLoc, -1, this->cmesh->noPoint); 
	for(int i = 0 ; i< this->cmesh->noPoint; i++)
	{
		if(i == iniPoint[0] || i == iniPoint[1] || i == iniPoint[2] || i == iniPoint[3]) continue;		
		int curTriId = this->c_locateIniTri(i, res); //walk to find triangle, -1 means inside.
		this->cmesh->pointMap[i] = curTriId;

		//if curTriId exists, register the point with maximum property
		if(curTriId != -1 && res >0) //then res == 0, the point is on the triangle
		{			
			if(res > this->triToValue[curTriId])
			{
				this->triToValue[curTriId] = res;
				this->triToPoint[curTriId] = i;
			}
			//set growing to be true
			this->growing = true;
            this->pointOrder[i] = i; 
            this->pointLoc[i] = curTriId; 
		}
	}

	int pos[5]; 

    c_reorderPoints( this->pointOrder, this->pointLoc, pos, 4, 0, this->cmesh->noPoint); 

    for ( int i = 0; i < 4; i++ ) 
        if ( pos[i] < pos[i+1] ) 
        {
            const StackItem item = { pos[i], pos[i+1], i }; 
            cStack.push_back(item); 
        }
}
void FFHull::c_exactInit()
{	
	exactinit();
}
int FFHull::c_locateIniTri(int pid, REAL & pred)
{
	pred = 0.0;
	int result = -1;
	REAL res30, res31, res32, res;
	int ip1, ip2, ip3;
	res31 = this->c_check3DOrientation(this->cmesh, -1, iniPoint[3], iniPoint[1], pid);
	res32 = this->c_check3DOrientation(this->cmesh, -1, iniPoint[3], iniPoint[2], pid);
	if((res31<0 && res32>0) || res31 == 0 || res32 == 0) //pick the smaller triangle
	{result = 1;ip1 = 1; ip2= 2;}
	else
	{		
		res30 = this->c_check3DOrientation(this->cmesh, -1, iniPoint[3], iniPoint[0], pid);
		if((res32<0 && res30>0) || res30 == 0)
		{result = 2;ip1 = 2; ip2= 0;}
		else if(res30<0 && res31>0)
		{result = 3;ip1 = 0; ip2= 1;}
	}
	res = this->c_check3DOrientation(this->cmesh, -1, iniPoint[ip1], iniPoint[ip2], pid);
	if(res >= 0) //below the plane, or on the boundary
	{result = 0; ip3 = Next(ip2);}
	else //above the plane
	{ip3 = ip2; ip2 = 3;}
	pred = this->c_check3DOrientation(this->cmesh, iniPoint[ip1], iniPoint[ip2], iniPoint[ip3], pid);
	if(pred <=0) //means inside the tetrahedral
		result = -1;
	return result;
}
void FFHull::c_expandArray(void ** data, int oldLength, int newLength)
{
	if(oldLength > newLength) 
	{
		printf("FFHull Error: Expand Array Invalid\n");
	}
	void * oldArray = * data;
	void * newArray;
	newArray = malloc(newLength);
	memcpy(newArray, oldArray, oldLength);
	*data = newArray;
	free(oldArray);
}

int FFHull::c_posInTriangle(Triangle * tri, int pid)
{
	int pos = -1;
	for(int i = 0; i< 3; i++)
	{
		if(tri->vtx[i] == pid)
		{
			pos = i;
			break;
		}
	}
	return pos;
}

void FFHull::c_clearTriangle(Triangle & tri)
{
	tri.vtx[0] = tri.vtx[1] = tri.vtx[2] = tri.neighbor[0] = tri.neighbor[1] = tri.neighbor[2] = -1;
}

//positive means i3 can see ccw-tri i0-i1-i2
REAL FFHull::c_check3DOrientation(TriangleMesh3D * tMesh, int i0, int i1, int i2, int i3)
{
	//1: Without SOS use our own logic, it is currently unstable
	/*Point3D p0, p1, p2, p3;
	p0 = (i0 == -1)?tMesh->kernel:tMesh->points[i0];
	p1 = (i1 == -1)?tMesh->kernel:tMesh->points[i1];
	p2 = (i2 == -1)?tMesh->kernel:tMesh->points[i2];
	p3 = (i3 == -1)?tMesh->kernel:tMesh->points[i3];
	return -orient3d((REAL *)&p0, (REAL *)&p1, (REAL*)&p2, (REAL*)&p3);	*/
	
	//2: Herbert's SOS provides a stable version
	return this->c_check3DOrientationSOS(tMesh, i0, i1, i2, i3);
}

REAL FFHull::c_check3DOrientationSOS(TriangleMesh3D * tMesh, int i0, int i1, int i2, int i3)
{
	Point3D v0, v1, v2, v3;
	v0 = (i0 == -1)?tMesh->kernel:tMesh->points[i0];
	v1 = (i1 == -1)?tMesh->kernel:tMesh->points[i1];
	v2 = (i2 == -1)?tMesh->kernel:tMesh->points[i2];
	v3 = (i3 == -1)?tMesh->kernel:tMesh->points[i3];

	int depth = 0, pn = 1;
	REAL p3[4][3];
	int index[4];
	REAL pa2[2], pb2[2], pc2[2];

	p3[0][0] = v0.x;
	p3[1][0] = v1.x;
	p3[2][0] = v2.x;
	p3[3][0] = v3.x;
	p3[0][1] = v0.y;
	p3[1][1] = v1.y;
	p3[2][1] = v2.y;
	p3[3][1] = v3.y;
	p3[0][2] = v0.z;
	p3[1][2] = v1.z;
	p3[2][2] = v2.z;
	p3[3][2] = v3.z;

	REAL result = -orient3d(p3[0], p3[1], p3[2], p3[3]);

	if(result == 0)
	{
		index[0] = i0; index[1] = i1; index[2] = i2; index[3] = i3;

        for (int i = 0; i < 3; i++) {
            int min = i; 

            for (int j = i + 1; j < 4; j++) 
                if (index[j] < index[min]) 
                    min = j; 

            if (min != i) {
                int tempI = index[min]; 
                index[min] = index[i]; 
                index[i] = tempI; 

                for (int m = 0; m < 3; m++) {
                    REAL tempR = p3[min][m]; 
                    p3[min][m] = p3[i][m]; 
                    p3[i][m] = tempR; 
                }

                pn = -pn; 
            }
        }

        //loop mode
		for(depth = 0; depth < 14; depth ++)
		{
			switch(depth)
			{
			case 0:
				pa2[0] = p3[1][0];
				pa2[1] = p3[1][1];
				pb2[0] = p3[2][0];
				pb2[1] = p3[2][1];
				pc2[0] = p3[3][0];
				pc2[1] = p3[3][1];
				break;
			case 1:
				pa2[0] = p3[1][0];
				pa2[1] = p3[1][2];
				pb2[0] = p3[2][0];
				pb2[1] = p3[2][2];
				pc2[0] = p3[3][0];
				pc2[1] = p3[3][2];
				break;
			case 2:
				pa2[0] = p3[1][1];
				pa2[1] = p3[1][2];
				pb2[0] = p3[2][1];
				pb2[1] = p3[2][2];
				pc2[0] = p3[3][1];
				pc2[1] = p3[3][2];
				break;
			case 3:
				pa2[0] = p3[0][0];
				pa2[1] = p3[0][1];
				pb2[0] = p3[2][0];
				pb2[1] = p3[2][1];
				pc2[0] = p3[3][0];
				pc2[1] = p3[3][1];
				break;
			case 4:
				result = p3[2][0] - p3[3][0];
				break;
			case 5:
				result = p3[2][1] - p3[3][1];
				break;
			case 6:
				pa2[0] = p3[0][0];
				pa2[1] = p3[0][2];
				pb2[0] = p3[2][0];
				pb2[1] = p3[2][2];
				pc2[0] = p3[3][0];
				pc2[1] = p3[3][2];
				break;
			case 7:
				result = p3[2][2] - p3[3][2];
				break;
			case 8:
				pa2[0] = p3[0][1];
				pa2[1] = p3[0][2];
				pb2[0] = p3[2][1];
				pb2[1] = p3[2][2];
				pc2[0] = p3[3][1];
				pc2[1] = p3[3][2];
				break;
			case 9:
				pa2[0] = p3[0][0];
				pa2[1] = p3[0][1];
				pb2[0] = p3[1][0];
				pb2[1] = p3[1][1];
				pc2[0] = p3[3][0];
				pc2[1] = p3[3][1];
				break;
			case 10:
				result = p3[1][0] - p3[3][0];
				break;
			case 11:
				result = p3[1][1] - p3[3][1];
				break;
			case 12:
				result = p3[0][0] - p3[3][0];
				break;
			default:
				result = 1.0;
				break;
			}

			if(depth < 4 || depth ==6 || depth == 8 || depth == 9)
				result = orient2d(pa2, pb2, pc2);
			if(result != 0)
			{
				if(depth == 1 || depth == 3 || depth == 5 || depth == 8 || depth == 10)
					result = - result;
				break;
			}
		}
	}

	return result * pn;
}

// Stable sort by pointLoc. Use counting sort. 
void FFHull::c_reorderPoints( int *pointOrder, char *pointLoc, int *offset, int noVal, 
                                   int beg, int end)
{
    int *newOrder = new int[end - beg + 1]; 
    int *counter = new int[noVal+1]; 

    for ( int i = 0; i < noVal + 1; i++ ) 
        counter[i] = 0; 

    for ( int i = beg; i < end; i++ ) 
        counter[pointLoc[i]+1]++; 

    offset[0] = 0; 
    offset[1] = 0; 

    for ( int i = 1; i < noVal; i++ )
        offset[i+1] = offset[i] + counter[i]; 

    for ( int i = beg; i < end; i++ ) 
    {
        int loc = pointLoc[i]; 

        if ( loc != -1 ) 
            newOrder[offset[loc+1]++] = pointOrder[i]; 
    }

    int newSize = offset[noVal]; 
    
    memcpy( pointOrder + beg, newOrder, newSize * sizeof(int) ); 

    delete [] newOrder; 
    delete [] counter; 
}
