/*
Author: Gao Mingcen & Cao Thanh Tung
Date: 28/02/2013

File Name: FFHull_CPU.cu

Functions of FFHull in the GPU version

===============================================================================

Copyright (c) 2012, 2013, School of Computing, National University of Singapore. 
All rights reserved.

Project homepage: http://www.comp.nus.edu.sg/~tants/flipflop.html

If you use ffHull and you like it or have comments on its usefulness etc., we 
would love to hear from you at <tants@comp.nus.edu.sg>. You may share with us
your experience and any possibilities that we may improve the work/code.

===============================================================================

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the National University of Singapore nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission from the National University of Singapore. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

#include <cuda_runtime.h>
#include <cutil.h>
#include <thrust/functional.h>
#include <thrust/remove.h>

#include "FFHull.h"
#include "FFHull_Kernel.h"

//thrust::function
struct Point3DToFloatXYZ : public thrust::unary_function<Point3D,float>
{
	__host__ __device__
	float operator()(Point3D p) { return (float)(p.x + p.y + p.z); }
};
struct Point3DToFloatX_Y_Z : public thrust::unary_function<Point3D,float>
{
	__host__ __device__
	float operator()(Point3D p) { return (float)(p.x - p.y - p.z); }
};
struct Point3DToFloat_XY_Z : public thrust::unary_function<Point3D,float>
{
	__host__ __device__
	float operator()(Point3D p) { return (float)(-p.x + p.y - p.z); }
};
struct Point3DToFloat_X_YZ : public thrust::unary_function<Point3D,float>
{
	__host__ __device__
	float operator()(Point3D p) { return (float)(-p.x - p.y + p.z); }
};
struct th_reportSpace : public thrust::unary_function<int,int>
{
__host__ __device__
int operator()(int triToPoint) { return triToPoint>=0?2:0; }
};

struct need_check
{
	__host__ __device__
	bool operator()(const int x)
	{
		return (x & 0x0000007)!=0;
	}
};
struct th_triSpace : public thrust::unary_function<int,int>
{
__host__ __device__
int operator()(Triangle tri) { return tri.vtx[0]!= -1?1:0;}
};
struct is_positive
{
	__host__ __device__
	bool operator()(const int x)
	{
		return x>=0;
	}
};
struct is_false
{
	__host__ __device__
	bool operator()(const bool x)
	{
		return !x;
	}
};

__global__ void cudaExactinit(REAL *constants)
{
    REAL half;
    REAL check, lastcheck;
    REAL cuda_epsilon = 1.0; 
    REAL cuda_splitter = 1.0; 
    int every_other;

    every_other = 1;
    half = 0.5;
    check = 1.0;
    // Repeatedly divide `cuda_epsilon' by two until it is too small to add to      
    //   one without causing roundoff.  (Also check if the sum is equal to     
    //   the previous sum, for machines that round up instead of using exact   
    //   rounding.  Not that these routines will work on such machines.)       
    do {
        lastcheck = check;
        cuda_epsilon *= half;
        if (every_other) {
            cuda_splitter *= 2.0;
        }
        every_other = !every_other;
        check = 1.0 + cuda_epsilon;
    } while ((check != 1.0) && (check != lastcheck));
    constants[0] = cuda_splitter + 1.0;											//splitter
    constants[1] = cuda_epsilon;												//epsilon											
    /* Error bounds for orientation and incircle tests. */
    constants[2] = (3.0 + 8.0 * cuda_epsilon) * cuda_epsilon;					//resulterrbound
    constants[3] = (3.0 + 16.0 * cuda_epsilon) * cuda_epsilon;					//ccwerrboundA
    constants[4] = (2.0 + 12.0 * cuda_epsilon) * cuda_epsilon;					//ccwerrboundB
    constants[5] = (9.0 + 64.0 * cuda_epsilon) * cuda_epsilon * cuda_epsilon;	//ccwerrboundC
    constants[6] = (10.0 + 96.0 * cuda_epsilon) * cuda_epsilon;					//iccerrboundA
    constants[7] = (4.0 + 48.0 * cuda_epsilon) * cuda_epsilon;					//iccerrboundB
    constants[8] = (44.0 + 576.0 * cuda_epsilon) * cuda_epsilon * cuda_epsilon;	//iccerrboundC
    constants[9] = (7.0 + 56.0 * cuda_epsilon) * cuda_epsilon;					//o3derrboundA
    constants[10] = (3.0 + 28.0 * cuda_epsilon) * cuda_epsilon;					//o3derrboundB
    constants[11] = (26.0 + 288.0 * cuda_epsilon) * cuda_epsilon * cuda_epsilon;//o3derrboundC
    constants[12] = 1.0 / sin(REAL(0.0));										// infinity
    constants[13] = (12.0 + 128.0 * cuda_epsilon) * cuda_epsilon;				// Weighted incircle
}

//host
void FFHull::g_exactInit()
{	
    REAL *d_constants;
    cutilSafeCall( cudaMalloc((void **)&d_constants, this->PredParaNum * sizeof(REAL)) ); 
    cudaExactinit<<<1,1>>>(d_constants); 
	cutilSafeCall(cudaMemcpy(this->hostConst, d_constants, this->PredParaNum * sizeof(REAL), cudaMemcpyDeviceToHost) ); 

	cutilSafeCall(cudaFree(d_constants));
}

void FFHull::GSetDevice(int id)
{
	cudaSetDevice(id);
}

void FFHull::g_Initialization()
{
	//exact computation
	this->g_exactInit(); 
	cutilSafeCall(cudaMemcpyToSymbol(deviceConst, hostConst, this->PredParaNum * sizeof(REAL)) ); 

	//manager
	this->primitiveManager = new GpuPrimitiveManager();
	this->memoryManager = new GpuMemoryManager(-1);

	//initialize gmesh and array
	gmesh = new TriangleMesh3D();
	gmesh->initialize();
	gmesh->noPoint = cmesh->noPoint;
	this->triArraySize = max(cmesh->noPoint/2, 4); //
	if(!this->memoryManager->GetMemory((void **) &gmesh->points, gmesh->noPoint * sizeof(Point3D)))exit(-1);
	if(!this->memoryManager->GetMemory((void **) &gmesh->pointMap, gmesh->noPoint * sizeof(int)))exit(-1);
	if(!this->memoryManager->GetMemory((void **) &g_pointIndex, gmesh->noPoint * sizeof(int)))exit(-1);	
	if(!this->memoryManager->GetMemory((void **) &gmesh->triangles, triArraySize * sizeof(Triangle)))exit(-1);
	if(!this->memoryManager->GetMemory((void **) &this->g_triPrefix, triArraySize * sizeof(int)))exit(-1);
	if(!this->memoryManager->GetMemory((void **) &this->triToPoint, triArraySize * sizeof(int)))exit(-1);
	if(!this->memoryManager->GetMemory((void **) &this->triToValue, triArraySize * sizeof(REAL)))exit(-1);

	cutilSafeCall(cudaMemcpy(gmesh->points, cmesh->points, gmesh->noPoint * sizeof(Point3D), cudaMemcpyHostToDevice));
	if(detail) CudaSync;
}
void FFHull::g_Deinitialization()
{
	this->g_copyMesh(this->cmesh);
	//free	
	if(!this->memoryManager->ReleaseMemory(this->g_triPrefix))exit(-1);
	if(!this->memoryManager->ReleaseMemory(this->triToValue))exit(-1);
	if(!this->memoryManager->ReleaseMemory(this->triToPoint))exit(-1);
	if(!this->memoryManager->ReleaseMemory(gmesh->triangles))exit(-1);
	if(!this->memoryManager->ReleaseMemory(gmesh->pointMap))exit(-1);
	if(!this->memoryManager->ReleaseMemory(gmesh->points))exit(-1);
	if(!this->memoryManager->ReleaseMemory(g_pointIndex))exit(-1);
	free(gmesh);

	this->primitiveManager->~GpuPrimitiveManager();
	this->memoryManager->~GpuMemoryManager();
	if(detail) CudaSync;
}
void FFHull::g_ComputeTetra()
{
	cudaFuncSetCacheConfig(cudaCorrectInitialGroupFast, cudaFuncCachePreferL1);

	int noPoint = gmesh->noPoint;
	float maxvalue[4]; //float is enough, we donot need the accurate
	//1.declare the array
	float * gValues = (float *) g_pointIndex; //if double , the length is not enough

	//2.compute value and the maximum:xyz, x_y_z, _xy_z, _x_yz	
	thrust::device_ptr<Point3D> in_pointer((Point3D *)gmesh->points);
	thrust::device_ptr<float> out_pointer((float *)gValues);
	thrust::transform(in_pointer, in_pointer+ noPoint, out_pointer, Point3DToFloatXYZ());
	this->iniPoint[0] = this->primitiveManager->maximumFloatElement(gValues, noPoint, maxvalue[0]);
	thrust::transform(in_pointer, in_pointer+ noPoint, out_pointer, Point3DToFloatX_Y_Z());
	this->iniPoint[1] = this->primitiveManager->maximumFloatElement(gValues, noPoint, maxvalue[1]);
	thrust::transform(in_pointer, in_pointer+ noPoint, out_pointer, Point3DToFloat_XY_Z());
	this->iniPoint[2] = this->primitiveManager->maximumFloatElement(gValues, noPoint, maxvalue[2]);
	thrust::transform(in_pointer, in_pointer+ noPoint, out_pointer, Point3DToFloat_X_YZ());
	this->iniPoint[3] = this->primitiveManager->maximumFloatElement(gValues, noPoint, maxvalue[3]);

	//2.5.if redudancy, random pick some...
	//3.compute the kernel
	cmesh->kernel.x = cmesh->kernel.y = cmesh->kernel.z = 0;
	for(int i =0; i< 4; i++)
	{
		cmesh->kernel.x += cmesh->points[iniPoint[i]].x;
		cmesh->kernel.y += cmesh->points[iniPoint[i]].y;
		cmesh->kernel.z += cmesh->points[iniPoint[i]].z;
	}
	cmesh->kernel.x/=4.0; gmesh->kernel.x = cmesh->kernel.x;
	cmesh->kernel.y/=4.0; gmesh->kernel.y = cmesh->kernel.y;
	cmesh->kernel.z/=4.0; gmesh->kernel.z = cmesh->kernel.z;

	//4. construct terahedral
	REAL res = this->c_check3DOrientation(cmesh, iniPoint[0], iniPoint[1], iniPoint[2], iniPoint[3]);
	if(res > 0)
	{
		int temp = iniPoint[0];
		iniPoint[0] = iniPoint[1];
		iniPoint[1] = temp;
	}
	gmesh->noTri = 4;
	Triangle cTris[4];
	cTris[0].vtx[0] = iniPoint[0];
	cTris[0].vtx[1] = iniPoint[1];
	cTris[0].vtx[2] = iniPoint[2];
	cTris[1].vtx[0] = iniPoint[2];
	cTris[1].vtx[1] = iniPoint[1];
	cTris[1].vtx[2] = iniPoint[3];
	cTris[2].vtx[0] = iniPoint[0];
	cTris[2].vtx[1] = iniPoint[2];
	cTris[2].vtx[2] = iniPoint[3];
	cTris[3].vtx[0] = iniPoint[1];
	cTris[3].vtx[1] = iniPoint[0];
	cTris[3].vtx[2] = iniPoint[3];
	cTris[0].neighbor[0] = Encode(1,2);
	cTris[0].neighbor[1] = Encode(2,2);
	cTris[0].neighbor[2] = Encode(3,2);
	cTris[1].neighbor[0] = Encode(3,1);
	cTris[1].neighbor[1] = Encode(2,0);
	cTris[1].neighbor[2] = Encode(0,0);
	cTris[2].neighbor[0] = Encode(1,1);
	cTris[2].neighbor[1] = Encode(3,0);
	cTris[2].neighbor[2] = Encode(0,1);
	cTris[3].neighbor[0] = Encode(2,1);
	cTris[3].neighbor[1] = Encode(1,0);
	cTris[3].neighbor[2] = Encode(0,2);
	cutilSafeCall(cudaMemcpy(gmesh->triangles, cTris, 4 * sizeof(Triangle), cudaMemcpyHostToDevice));

	//5. initialize triToPoint, triToValue
	int cTTP[4]={-1, -1, -1, -1};
	REAL cTTV[4] = {MINREAL,MINREAL,MINREAL,MINREAL};
	cutilSafeCall(cudaMemcpy(this->triToPoint, cTTP, 4 * sizeof(int), cudaMemcpyHostToDevice));
	cutilSafeCall(cudaMemcpy(this->triToValue, cTTV, 4 * sizeof(REAL), cudaMemcpyHostToDevice));
	
	//5. Walking to find the correct group
	int threadNum = 128;
	int blockNum = GET_GRID(noPoint, threadNum);
	//bool isExact = true;
	//cutilSafeCall(cudaMemcpy(this->g_exact, &isExact, sizeof(bool), cudaMemcpyHostToDevice));
	cudaCorrectInitialGroupFast<<<blockNum, threadNum>>>(gmesh->points, gmesh->pointMap, gmesh->kernel,
		this->triToPoint, this->triToValue, noPoint, iniPoint[0], iniPoint[1], iniPoint[2], iniPoint[3]);
	cudaCorrectInitialGroupExact<<<blockNum, threadNum>>>(gmesh->points, gmesh->pointMap, gmesh->kernel,
		this->triToPoint, this->triToValue, noPoint, iniPoint[0], iniPoint[1], iniPoint[2], iniPoint[3]);
	if(detail) cutilCheckError();

	//6. Build the index of alive point based on pointMap information
	g_noAlive = this->primitiveManager->signCompactPos(g_pointIndex, gmesh->pointMap, noPoint);
	this->growing = (g_noAlive>0);

	//7. set the map for initial points
	for(int i = 0; i< 4; i++)
		cutilSafeCall(cudaMemcpy(gmesh->pointMap + this->iniPoint[i], &i, sizeof(int), cudaMemcpyHostToDevice));
	if(detail) CudaSync;
}
void FFHull::g_GrowHull()
{
	cudaFuncSetCacheConfig(cudaSubdividePoints, cudaFuncCachePreferL1);
	int count = 0;
	if(growing)
	{
		count ++;
		this->growing = false;
		//1. Assign the layer
		this->g_assignLayer();
		//2. Subdivide the point, and set growing to be true if necessary
		this->g_subdivide();
	}
	if(detail) CudaSync;
}

void FFHull::g_CorrectHull()
{
	//1. initialization
	//the formate in triCheckingStatus OxFFFFFFFF
	//the loweast 8 bits: 000-111, meaning which edge need to check. 1: to check. 0: no need
	//the other three 8 bits: 00/01/11, meaning the status of edge. 
	//000: haven't check or no action. 
	//100: 3-1 flipping(Next reflex). 101: 3-1 flipping(Prev reflex). 110: 2-2 flipping. 
	//010: non extreme point(Next reflex). 011: non extreme point(Prev reflex).
	int count = 0;
	int noTri = gmesh->noTri;
	int * triStatus = this->triToPoint;
	thrust::device_ptr<int> status_pointer(triStatus);
	int * triIndex = (int *) this->triToValue;
	thrust::device_ptr<int> index_pointer(triIndex);
	int * triLabels;
	if(!this->memoryManager->GetMemory((void **) &triLabels, noTri * sizeof(int)))exit(-1);
	thrust::device_ptr<int> label_pointer(triLabels);
	thrust::counting_iterator<int> pos_pointer(0);
	thrust::fill(status_pointer, status_pointer + noTri, 0x00000007);
	thrust::fill(label_pointer, label_pointer + noTri, MAXINT);

	//2. enter the loop
	int checkNum;
	bool check = true;
	bool * g_check;
	if(!this->memoryManager->GetMemory((void **) &g_check, sizeof(bool)))exit(-1);

	while(check)
	{
		count++;
		check=false;
		cutilSafeCall(cudaMemcpy(g_check, & check, sizeof(bool), cudaMemcpyHostToDevice));

		//2.1 compact to get triIndex, which is used for checking, flipping and updating
		thrust::device_ptr<int> comp_pointer = 
			thrust::copy_if(pos_pointer, pos_pointer + noTri, status_pointer, index_pointer, need_check());
		checkNum = comp_pointer - index_pointer;
		//if(checkNum == 0) break;

		//2.2 check the edge
		//if the edge bit is 1 ,check and label in status of the edge pos. When one edge needs to flip, break
		//never change 3 edge bits
		int threadNum = 128;
		int blockNum = GET_GRID(checkNum, threadNum);
		cudaCheckTriangles<<<blockNum, threadNum>>>(triIndex, triStatus, triLabels, gmesh->triangles, gmesh->points, 
			gmesh->pointMap, gmesh->kernel, noTri, checkNum);
		if(detail) cutilCheckError();

		//2.3 fill the edge
		//find the first edge bit =1, read the code, 
		//see whether the triangle lable of induced sub-complex are equal to it. If it is flip. Fill it self information
		cudaPrepareFlipping<<<blockNum, threadNum>>>(triIndex, triStatus, triLabels, gmesh->triangles, noTri, checkNum);
		if(detail) cutilCheckError();
		cudaFlipTriangles<<<blockNum, threadNum>>>(triIndex, triStatus, triLabels, gmesh->triangles, 
			gmesh->pointMap, noTri, checkNum);
		if(detail) cutilCheckError();

		//2.4 repair the topology
		cudaFixTopology<<<blockNum, threadNum>>>(triIndex, triStatus, triLabels, gmesh->triangles, checkNum);
		if(detail) cutilCheckError();

		//2.5 clearLabel
		cudaClearLabel<<<blockNum, threadNum>>>(triIndex, triStatus,triLabels, gmesh->triangles, g_check, checkNum);
		if(detail) cutilCheckError();
		cutilSafeCall(cudaMemcpy(& check, g_check, sizeof(bool), cudaMemcpyDeviceToHost));
	}
	if(!this->memoryManager->ReleaseMemory(g_check)) exit(-1);
	if(!this->memoryManager->ReleaseMemory(triLabels)) exit(-1);
	if(detail) printf("**** Flip triangles by %d rounds\n", count);
	if(detail) CudaSync;
}

void FFHull::g_CompactMesh()
{
	//1. report 0, 1 array
	int oldNoTri = gmesh->noTri;
	int * spaces = this->triToPoint;	
	thrust::device_ptr<Triangle> in_pointer(gmesh->triangles);
	thrust::device_ptr<int> out_pointer(spaces);
	thrust::transform(in_pointer, in_pointer + oldNoTri, out_pointer, th_triSpace());

	//2. prefix sum
	int newNoTri, lastNum;
	cutilSafeCall(cudaMemcpy(&lastNum, spaces + oldNoTri -1, sizeof(int), cudaMemcpyDeviceToHost));
	this->primitiveManager->exclusivePrefixSum(spaces, spaces, oldNoTri);
	cutilSafeCall(cudaMemcpy(&newNoTri, spaces + oldNoTri -1, sizeof(int), cudaMemcpyDeviceToHost));
	newNoTri += lastNum; //newNoTri is the number of new created ones, not the total one

	if(newNoTri != oldNoTri)
	{
		//3. launch a kernel to compact
		int threadNum = 128;
		int blockNum = GET_GRID(oldNoTri, threadNum);
		Triangle * newTriangles;
		if(!this->memoryManager->GetMemory((void **) &newTriangles, newNoTri * sizeof(Triangle)))exit(-1);	

		cudaCompactTriangles<<<blockNum, threadNum>>>(gmesh->triangles, newTriangles, spaces, oldNoTri);
		if(detail) cutilCheckError();
		
		if(!this->memoryManager->ReleaseMemory(gmesh->triangles)) exit(-1);
		gmesh->triangles = newTriangles;
		gmesh->noTri = newNoTri;

		//4. update pointMap
		blockNum = GET_GRID(gmesh->noPoint, threadNum);
		cudaUpdatePointMap<<<blockNum, threadNum>>>(gmesh->pointMap, spaces, gmesh->noPoint);
		if(detail) cutilCheckError();
	}
	if(detail) printf("**** Compact triangles from %d to %d\n", oldNoTri, this->gmesh->noTri);
	if(detail) CudaSync;
}
void FFHull::g_assignLayer()
{	
	int oldNoTri = this->gmesh->noTri;
	
	//1. report the space for each triangle	
	thrust::device_ptr<int> in_pointer(this->triToPoint);
	thrust::device_ptr<int> out_pointer(this->g_triPrefix);
	thrust::transform(in_pointer, in_pointer + oldNoTri, out_pointer, th_reportSpace());
	
	//2. prefix the total, if needed expand the space
	int newNoTri, lastNum;
	cutilSafeCall(cudaMemcpy(&lastNum, this->g_triPrefix + oldNoTri -1, sizeof(int), cudaMemcpyDeviceToHost));
	this->primitiveManager->exclusivePrefixSum(this->g_triPrefix, this->g_triPrefix, oldNoTri);
	cutilSafeCall(cudaMemcpy(&newNoTri, this->g_triPrefix + oldNoTri -1, sizeof(int), cudaMemcpyDeviceToHost));
	newNoTri += lastNum; //newNoTri is the number of new created ones, not the total one

	if(newNoTri + oldNoTri > triArraySize)
	{
		//gMesh->triangles, this->triToPoint, this->triToValue
		triArraySize = (newNoTri + oldNoTri) * 2;
		this->memoryManager->ExpandArray((void **) &gmesh->triangles, oldNoTri * sizeof(Triangle), triArraySize * sizeof(Triangle));
		this->memoryManager->ExpandArray((void **) &this->triToPoint, oldNoTri * sizeof(int), triArraySize * sizeof(int));
		this->memoryManager->ExpandArray((void **) &this->triToValue, oldNoTri * sizeof(REAL), triArraySize * sizeof(REAL));		
		this->memoryManager->ExpandArray((void **) &this->g_triPrefix, oldNoTri * sizeof(int), triArraySize * sizeof(int));
	}
	
	//3. expand triangle
	//the rule is: for one triangle, find the offset in prefix, plus oldNoTri, is the beginning space for it. 
	//The neighbors inside the old triangle can be filled. 
	//several things to do:
	//	a. expand the triangles, reset triToPoint and triToValue
	//	b. refresh pointMap

	int threadNum = 128;
	int blockNum = GET_GRID(oldNoTri, threadNum);
	cudaExpandTriangles<<<blockNum, threadNum>>>(gmesh->triangles, gmesh->pointMap, 
		this->triToPoint, this->triToValue, this->g_triPrefix, oldNoTri);
	if(detail) cutilCheckError();
	
	//4. register checked edge
	//several things to do
	//	a. fill the neighbor between boundary
	//  b. fill checked edges
	//the reason why we cannot merge this in step 3: we cannot find the correct pair
	blockNum = (newNoTri/2 - 1) / threadNum + 1;	
	blockNum = GET_GRID(newNoTri/2, threadNum);
	cudaFixBoundary<<<blockNum, threadNum>>>(gmesh->triangles, this->g_triPrefix, oldNoTri, newNoTri/2);
	if(detail) cutilCheckError();

	gmesh->noTri = oldNoTri + newNoTri;
	if(detail) cutilCheckError();
}
void FFHull::g_subdivide()
{	
	//check each point, if it is not on the triangle, select a sub-devided triangle. 
	//Fill pointMap, triToPoint and triToValue
	int threadNum = 128;
	int blockNum = GET_GRID(g_noAlive, threadNum);
	cudaSubdividePoints<<<blockNum, threadNum>>>(gmesh->triangles, gmesh->points, gmesh->pointMap, this->g_pointIndex, 
		this->triToPoint, this->triToValue, gmesh->kernel, g_noAlive);
	if(detail) cutilCheckError(); 

	//shrink the index
	bool * isAliveArray;
	if(!this->memoryManager->GetMemory((void **) &isAliveArray, g_noAlive * sizeof(bool)))exit(-1);
	blockNum = GET_GRID(g_noAlive, threadNum);
	cudaLabelAlivePoints<<<blockNum, threadNum>>>(gmesh->triangles, gmesh->pointMap, this->g_pointIndex, isAliveArray, g_noAlive);
	if(detail) cutilCheckError();
	
	g_noAlive = this->primitiveManager->removeFalse(g_pointIndex, isAliveArray, g_noAlive); 
	this->growing = (g_noAlive>0);
	
	if(!this->memoryManager->ReleaseMemory(isAliveArray))exit(-1);
}
void FFHull::g_copyMesh(TriangleMesh3D * host_mesh)
{
	//copy from gmesh to cmesh
	host_mesh->noTri = gmesh->noTri;
	host_mesh->triangles = (Triangle *) malloc(host_mesh->noTri * sizeof(Triangle));
	cutilSafeCall(cudaMemcpy(host_mesh->pointMap, gmesh->pointMap, host_mesh->noPoint * sizeof(int), cudaMemcpyDeviceToHost));
	cutilSafeCall(cudaMemcpy(host_mesh->triangles, gmesh->triangles, host_mesh->noTri * sizeof(Triangle), cudaMemcpyDeviceToHost));
}