/*
Author: Gao Mingcen
Date: 28/02/2013

File Name: GpuPrimitiveManager.cu

Functions of GpuPrimitiveManager, a simple manager of GPU primitives using Thrust

===============================================================================

Copyright (c) 2012, 2013, School of Computing, National University of Singapore. 
All rights reserved.

Project homepage: http://www.comp.nus.edu.sg/~tants/flipflop.html

If you use ffHull and you like it or have comments on its usefulness etc., we 
would love to hear from you at <tants@comp.nus.edu.sg>. You may share with us
your experience and any possibilities that we may improve the work/code.

===============================================================================

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the National University of Singapore nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission from the National University of Singapore. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

#include <cuda_runtime.h>
#include <cutil.h>
#include <thrust/remove.h>
#include "GpuPrimitiveManager.h"

struct is_positive
{
	__host__ __device__
	bool operator()(const int x)
	{
		return x>=0;
	}
};
struct is_false
{
	__host__ __device__
	bool operator()(const bool x)
	{
		return !x;
	}
};
GpuPrimitiveManager::GpuPrimitiveManager()
{
}

GpuPrimitiveManager::~GpuPrimitiveManager(void)
{
}

void GpuPrimitiveManager::exclusivePrefixSum(int * out, int * in, int size)
{	
	thrust::device_ptr<int> in_pointer((int *)in);
	thrust::device_ptr<int> out_pointer((int *)out);
	thrust::exclusive_scan(in_pointer, in_pointer + size, out_pointer); 
}

int GpuPrimitiveManager::signCompactPos(int * out, int * inValid, int size)
{
	thrust::counting_iterator<int> pos_pointer(0);	
	thrust::device_ptr<int> out_pointer((int *)out);	
	thrust::device_ptr<int> stencil_pointer((int *)inValid);
	thrust::device_ptr<int> final_pointer = 
		thrust::copy_if(pos_pointer, pos_pointer + size, stencil_pointer, out_pointer, is_positive());
	int num = final_pointer - out_pointer;
	return num;
}

int GpuPrimitiveManager::removeFalse(int * in, bool * isValid, int size)
{
	thrust::device_ptr<int> in_pointer((int *)in);
	thrust::device_ptr<bool> stencil_pointer((bool *) isValid);
	thrust::device_ptr<int> final_pointer 
		= thrust::remove_if(in_pointer, in_pointer + size, stencil_pointer, is_false());
	int num = final_pointer - in_pointer;
	return num;
}

//return the offset of the maximum; value store the maxmum float value
int GpuPrimitiveManager::maximumFloatElement(float * in, size_t size, float & value)
{
	thrust::device_ptr<float> in_pointer((float *) in);
	thrust::device_ptr<float> max_pointer = thrust::max_element(in_pointer, in_pointer+size);
	value = *max_pointer;
	return max_pointer - in_pointer;
}