/*
Author: Gao Mingcen & Cao Thanh Tung
Date: 28/02/2013

File Name: common.h

Data struction and macro defintion for the projects

===============================================================================

Copyright (c) 2012, 2013, School of Computing, National University of Singapore. 
All rights reserved.

Project homepage: http://www.comp.nus.edu.sg/~tants/flipflop.html

If you use ffHull and you like it or have comments on its usefulness etc., we 
would love to hear from you at <tants@comp.nus.edu.sg>. You may share with us
your experience and any possibilities that we may improve the work/code.

===============================================================================

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the National University of Singapore nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission from the National University of Singapore. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

#ifndef __COMMON_H__
#define __COMMON_H__

#include <cuda_runtime.h>
#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include <map>
#include "PerformanceTimer.h"

using namespace std;

#define EXACT                           // Turns on/off exact arithmetics
#define SINGLE_PRECISION
typedef unsigned int uint; 
#ifdef SINGLE_PRECISION
#define REAL float
#else
#define REAL double
#endif

#define MINREAL -1E10
#define MAXREAL 1E10
#define MAXINT 0x7FFFFFFF
#define MININT -0x7FFFFFFF
#define PI 3.141592654

//triangle
#define Next(p) ((p)+1)%3
#define Prev(p) ((p)+2)%3
#define Encode(t,p) (((t)<<2)|(p))
#define Pos(c) ((c)&3)
#define Tri(c) ((c)>>2)

//timing
#define DefineClock PerformanceTimer timer
#define BeginClock	timer.StartTimer()
#define EndClock	printf("Computing time: %.6f ms\n",timer.GetTimeElapsed()*1000)
#define GetClock	timer.GetTimeElapsed()*1000
#define CudaSync	cudaThreadSynchronize()

//grid size
#define MAXBLOCKS   16384
#define GET_GRID(items, block)      min((items-1) / (block) + 1, MAXBLOCKS)

//Cuda checking
#define CUDA_CHECK_ERROR

#define cutilSafeCall(err)           __cudaSafeCall      (err, __FILE__, __LINE__)
#define cutilCheckError()            __cudaCheckError   (__FILE__, __LINE__)
inline void __cudaSafeCall( cudaError err, const char *file, const int line )
{
#ifdef CUDA_CHECK_ERROR
    do {
        if( cudaSuccess != err) {
            fprintf(stderr, "cudaSafeCall() Runtime API error in file <%s>, line %i : %s.\n",
                    file, line, cudaGetErrorString( err) );
            exit(-1);
        }
    } while (0);
#endif
}

inline void __cudaCheckError( const char *file, const int line )
{
#ifdef CUDA_CHECK_ERROR
    do {
        cudaError_t err = cudaGetLastError();
        if( cudaSuccess != err) {
            fprintf(stderr, "cutilCheckMsg() CUTIL CUDA error in file <%s>, line %i : %s.\n",
                    file, line, cudaGetErrorString( err) );
            exit(-1);
        }
		// More careful checking will be to perform a cudaThreadSynchronize(). 
		// This however will reduce the performance. 
        err = cudaThreadSynchronize();
        if( cudaSuccess != err) {
            fprintf(stderr, "cutilCheckMsg cudaThreadSynchronize error in file <%s>, line %i : %s.\n",
                    file, line, cudaGetErrorString( err) );
            exit(-1);
        }
    } while (0);
#endif
}

//data structure
struct Point3D
{
	REAL x, y, z;
};

struct Edge
{
	int a,b;
	Edge(){};
	Edge(int _a, int _b)
	{
		this->a = _a;
		this->b = _b;
	};
};

struct Triangle
{
	unsigned int vtx[3];
	unsigned int neighbor[3];
};

struct TriangleMesh3D
{
	int noPoint;
	int noTri;
	Point3D * points;
	Triangle * triangles;
	int * pointMap;	//map to one of the triangles containing the point
	Point3D kernel;

	TriangleMesh3D()
	{
		this->initialize();
	};
	void initialize()
	{
		this->noPoint = 0;
		this->noTri = 0; 
		this->points = NULL;
		this->triangles = NULL;
		this->pointMap = NULL;
	};
	void initialize(int seednum)
	{
		this->noPoint = seednum;
		this->points = (Point3D *) malloc(this->noPoint * sizeof(Point3D));
		this->pointMap = (int *) malloc(this->noPoint * sizeof(int));
		memset(this->pointMap, -1, this->noPoint * sizeof(int));
	};
	void clear()
	{
		this->noPoint = 0;this->noTri = 0; 
		if(this->points != NULL)
		{free(this->points);this->points = NULL;}
		if(this->triangles != NULL)
		{free(this->triangles);this->triangles = NULL;}
		if(this->pointMap != NULL)
		{free(this->pointMap);this->pointMap = NULL;}
	};
};

struct GpuMemoryUnit
{
	int * pointer;
	int byteLength;
};

#endif